@@ -2,11 +2,11 @@
 """
 from collections import defaultdict
 import xarray
-from xarray.core.utils import V
 from .template import Template, Placeholders
 from typing import Generator, Optional, Dict, Any, Sequence, Union
 from pathlib import Path
 import numpy as np
+from functools import cmp_to_key
 
 
 class FileTree:
@@ -15,16 +15,16 @@ class FileTree:
     The many methods can be split into 4 categories
 
         1. The template interface. Each path (file or directory) is reprsented by a :class:`Template <file_tree.template.Template>`, 
-            which defines the filename with any unknown parts (e.g., subject ID) marked by placeholders. 
-            Templates are accessed based on their key.
+           which defines the filename with any unknown parts (e.g., subject ID) marked by placeholders. 
+           Templates are accessed based on their key.
 
             - :meth:`get_template`: used to access a template based on its key.
-            - :meth:`add_template`: used to add a new template or overwrite an existing one.
             - :meth:`template_keys`: used to list all the template keys.
-            - :meth:`add_subtree`: can be used to add all the templates from a different tree to the present one.
+            - :meth:`add_template`: used to add a new template or overwrite an existing one.
+            - :meth:`add_subtree`: can be used to add all the templates from a different tree to this one.
         
         2. The placeholder interface. Placeholders represent values to be filled into the placeholders.
-            Each placeholder can be either undefined, have a singular value, or have a sequence of possible values.
+           Each placeholder can be either undefined, have a singular value, or have a sequence of possible values.
 
             - You can access the :class:`placeholders dictionary-like object <file_tree.template.Placeholders>` directly through `FileTree.placeholders`
             - :meth:`update`: returns a new FileTree with updated placeholders or updates the placeholders in the current one.
@@ -35,9 +35,12 @@ class FileTree:
         3. Getting the actual filenames based on filling the placeholder values into the templates.
 
             - :meth:`get`: Returns a valid path by filling in all the placeholders in a template. 
-                For this to work all placeholder values should be defined and singular.
+              For this to work all placeholder values should be defined and singular.
             - :meth:`get_mult`: Returns array of all possible valid paths by filling in the placeholders in a template.
-                Placeholder values can be singular or a sequence of possible values.
+              Placeholder values can be singular or a sequence of possible values.
+            - :meth:`get_mult_glob`: Returns array with existing paths on disk. 
+              Placeholder values can be singular, a sequence of possible values, or undefined.
+              In the latter case possible values for that placeholder are determined by checking the disk.
             - :meth:`fill`: Returns new FileTree with any singular values filled into the templates and removed from the placeholder dict.
 
         4. Input/output
@@ -49,13 +52,14 @@ class FileTree:
             - :meth:`to_string`: writes a FileTree to a string.
     """
 
-    def __init__(self, templates: Dict[str, Template], placeholders: Dict[str, Any]):
+    def __init__(self, templates: Dict[str, Template], placeholders: Dict[str, Any], return_path=False):
         self._templates = templates
         self.placeholders = Placeholders(placeholders)
+        self.return_path = return_path
 
     # create new FileTree objects
     @classmethod
-    def empty(cls, top_level: Union[str, Template]='.') -> "FileTree":
+    def empty(cls, top_level: Union[str, Template]='.', return_path=False) -> "FileTree":
         """Creates a new empty FileTree containing only a top-level directory
 
         Args:
@@ -66,10 +70,10 @@ class FileTree:
         """
         if not isinstance(top_level, Template):
             top_level = Template(None, top_level)
-        return cls({"": top_level}, {})
+        return cls({"": top_level}, {}, return_path=return_path)
 
     @classmethod
-    def read(cls, name: str, top_level: Union[str, Template]='.', **placeholders) -> "FileTree":
+    def read(cls, name: str, top_level: Union[str, Template]='.', return_path=False, **placeholders) -> "FileTree":
         """Reads a filetree based on the given name
 
         Args:
@@ -91,12 +95,12 @@ class FileTree:
                 with open(filename, 'r') as f:
                     text = f.read()
                 with parse_tree.extra_tree_dirs([Path(filename).parent]):
-                    return cls.from_string(text, top_level, **placeholders)
+                    return cls.from_string(text, top_level, return_path=return_path, **placeholders)
         text = parse_tree.search_tree(name)
-        return cls.from_string(text, top_level, **placeholders)
+        return cls.from_string(text, top_level, return_path=return_path, **placeholders)
 
     @classmethod
-    def from_string(cls, definition: str, top_level: Union[str, Template]='.', **placeholders) -> "FileTree":
+    def from_string(cls, definition: str, top_level: Union[str, Template]='.', return_path=False, **placeholders) -> "FileTree":
         """Creates a FileTree based on the given definition
 
         Args:
@@ -107,7 +111,9 @@ class FileTree:
             FileTree: tree matching the definition in the file
         """
         from . import parse_tree
-        return parse_tree.read_file_tree_text(definition.splitlines(), top_level).update(inplace=True, **placeholders)
+        res = parse_tree.read_file_tree_text(definition.splitlines(), top_level).update(inplace=True, **placeholders)
+        res.return_path = return_path
+        return res
 
     def copy(self, ) -> "FileTree":
         """Creates a copy of the tree
@@ -117,7 +123,7 @@ class FileTree:
         Returns:
             FileTree: new tree object with identical templates, sub-trees and placeholders
         """
-        new_tree = type(self)(dict(self._templates), Placeholders(self.placeholders))
+        new_tree = type(self)(dict(self._templates), Placeholders(self.placeholders), self.return_path)
         return new_tree
 
     # template interface
@@ -137,32 +143,35 @@ class FileTree:
 
     @property
     def top_level(self, ):
-        return Path(self.get_template("").unique_part)
+        as_string = self.get_template("").unique_part
+        if self.return_path:
+            return Path(as_string)
+        return str(as_string)
 
     @top_level.setter
     def top_level(self, value: str):
         self.get_template("").unique_part = str(value)
 
-    def add_template(self, template_path: str, key: Optional[Union[str, Sequence[str]]]=None, parent_key: str="", overwrite=False) -> str:
+    def add_template(self, template_path: str, key: Optional[Union[str, Sequence[str]]]=None, parent: str="", overwrite=False) -> str:
         """Updates the FileTree with the new template
 
         Args:
             template_path: path name with respect to the parent (or top-level if no parent provided)
             key: key(s) to access this template in the future. Defaults to result from :meth:`Template.guess_key <file_tree.template.Template.guess_key>` 
                 (i.e., the path basename without the extension).
-            parent_key: if defined, `template_path` will be interpreted as relative to this template. 
+            parent: if defined, `template_path` will be interpreted as relative to this template. 
                 By default the top-level template is used as reference.
-                To create a template unaffiliated with the rest of the tree, set `parent_key` to None. 
+                To create a template unaffiliated with the rest of the tree, set `parent` to None. 
                 Such a template should be an absolute path or relative to the current directory and can be used as parent for other templates
             overwrite: if True, overwrites any existing template rather than raising a ValueError. Defaults to False.
 
         Returns:
             str: one of the short names under which the template has been stored
         """
-        parent = None if parent_key is None else self.get_template(parent_key)
-        new_template = Template(parent, template_path)
-        if parent is not None and new_template.as_path == parent.as_path:
-            new_template = parent
+        parent_template = None if parent is None else self.get_template(parent)
+        new_template = Template(parent_template, template_path)
+        if parent_template is not None and new_template.as_path == parent_template.as_path:
+            new_template = parent_template
         return self._add_actual_template(new_template, key, overwrite=overwrite)
 
     def _add_actual_template(self, template: Template, keys: Optional[Union[str, Sequence[str]]]=None, overwrite=False):
@@ -183,17 +192,23 @@ class FileTree:
             self._templates[key] = template
         return keys[0]
 
-    def template_keys(self, ):
+    def template_keys(self, only_leaves=False):
         """Returns the keys of all the templates in the FileTree
 
         Each key will be returned for templates with multiple keys.
-        """
-        return self._templates.keys()
 
-    def add_subtree(self, sub_tree: "FileTree", precursor: Optional[Union[str, Sequence[str]]]=(None, ), parent_template: Optional[str]="") -> None:
+        Args
+            only_leaves (bool, optional): set to True to only return templates that do not have any children
+        """
+        if not only_leaves:
+            return self._templates.keys()
+        parents = {t.parent for t in self._templates.values() if t.parent is not None}
+        return {key for key, template in self._templates.items() if template not in parents}
+
+    def add_subtree(self, sub_tree: "FileTree", precursor: Optional[Union[str, Sequence[str]]]=(None, ), parent: Optional[str]="") -> None:
         """Updates the templates and the placeholders in place with those in sub_tree
 
-        The top-level directory of the sub-tree will be replaced by the `parent_template` (unless set to None).
+        The top-level directory of the sub-tree will be replaced by the `parent` (unless set to None).
         The sub-tree templates will be available with the key "<precursor>/<original_key>",
         unless the precursor is None in which case they will be unchanged (which can easily lead to errors due to naming conflicts).
 
@@ -213,7 +228,9 @@ class FileTree:
         Args:
             sub_tree (FileTree): tree to be added to the current one
             precursor (list(str or None)): name(s) of the sub-tree. Defaults to just adding the sub-tree to the main tree without precursor
-            parent_template (str): key of the parent template. Defaults to top-level directory. Can be set to None for an independent tree.
+            parent (str): key of the template used as top-level directory for the sub tree. 
+                Defaults to top-level directory of the main tree. 
+                Can be set to None for an independent tree.
         """
         for name in precursor:
             if name is None:
@@ -226,21 +243,21 @@ class FileTree:
             to_assign = dict(sub_tree_fill._templates)
             assigned = {"": (
                 Template(None, to_assign[""].unique_part) 
-                if parent_template is None else 
-                self.get_template(parent_template)
+                if parent is None else 
+                self.get_template(parent)
             )}
             been_assigned = {to_assign[""]: ""}
             del to_assign[""]
             while len(to_assign) > 0:
                 for key, old_template in list(to_assign.items()):
                     if old_template.parent is None:
-                        parent = None
+                        parent_template = None
                     elif old_template.parent in been_assigned:
                         parent_key = been_assigned[old_template.parent]
-                        parent = assigned[parent_key]
+                        parent_template = assigned[parent_key]
                     else:
                         continue
-                    assigned[key] = Template(parent, old_template.unique_part).add_precursor(add_string)
+                    assigned[key] = Template(parent_template, old_template.unique_part).add_precursor(add_string)
                     been_assigned[old_template] = key
                     del to_assign[key]
 
@@ -258,8 +275,11 @@ class FileTree:
                 conflict = {key for key in sub_tree_fill.placeholders.keys() if key in self.placeholders}
                 if len(conflict) > 0:
                     raise ValueError(f"Sub-tree placeholder values for {conflict} conflict with those set in the parent tree.")
-            self.placeholders.update({add_string + key: value for key, value in sub_tree_fill.placeholders.items()})
-
+            for old_key, old_value in sub_tree_fill.placeholders.items():
+                if isinstance(old_key, str):
+                    self.placeholders[add_string + old_key] = old_value
+                else:
+                    self.placeholders[frozenset(add_string + k for k in old_key)] = {add_string + k: v for k, v in old_value.items()}
 
     # placeholders interface
     def update(self, inplace=False, **placeholders) -> "FileTree":
@@ -277,11 +297,11 @@ class FileTree:
         new_tree.placeholders.update(placeholders)
         return new_tree
 
-    def update_glob(self, template_key: str, inplace=False) -> "FileTree":
-        """Updates any undefined placeholders based on the which files exist on disk for template
+    def update_glob(self, template_key: Union[str, Sequence[str]], inplace=False) -> "FileTree":
+        """Updates any undefined placeholders based on which files exist on disk for template
 
         Args:
-            template_key (str): key of the template to use
+            template_key (str or sequence of str): key(s) of the template(s) to use
             inplace (bool): if True change the placeholders in-place (and return the FileTree itself); 
                 by default a new FileTree is returned with the updated values without altering this one.
 
@@ -296,10 +316,21 @@ class FileTree:
             from_template = template.get_all_placeholders(self.placeholders)
             for name, values in from_template.items():
                 new_placeholders[name] = new_placeholders[name].union(values)
-        return self.update(inplace, **new_placeholders)
+
+        def cmp(item1, item2):
+            if item1 is None:
+                return -1
+            if item2 is None:
+                return 1
+            if item1 < item2:
+                return -1
+            if item1 > item2:
+                return 1
+            return 0
+        return self.update(inplace, **{k: sorted(v, key=cmp_to_key(cmp)) for k, v in new_placeholders.items()})
 
     # Extract paths
-    def get(self, key: str, make_dir=False) -> Path:
+    def get(self, key: str, make_dir=False) -> Union[str, Path]:
         """Returns template with placeholder values filled in
 
         Args:
@@ -307,44 +338,95 @@ class FileTree:
             make_dir (bool, optional): If set to True, create the parent directory of the returned path.
 
         Returns:
-            Path: Filled in template as Path object
+            Path: Filled in template as Path object. 
+                Returned as a `pathlib.Path` object if `FileTree.return_path` is True. 
+                Otherwise a string is returned.
         """
         path = self.get_template(key).format_single(self.placeholders)
         if make_dir:
             path.parent.mkdir(parents=True, exist_ok=True)
+        if self.return_path:
+            return Path(path)
         return path
 
-    def get_mult(self, key: Union[str, Sequence[str]], make_dir=False) -> Union[xarray.DataArray, xarray.Dataset]:
+    def get_mult(self, key: Union[str, Sequence[str]], filter=False, make_dir=False) -> Union[xarray.DataArray, xarray.Dataset]:
         """Returns array of paths with all possible values filled in for the placeholders 
 
         Singular placeholder values are filled into the template directly. 
         For each placeholder with multiple values a dimension is added to the output array.
         This dimension will have the name of the placeholder and labels corresponding to the possible values (see http://xarray.pydata.org/en/stable/).
+        The precense of required, undefined placeholders will lead to an error 
+        (see :meth:`get_mult_glob` or :meth:`update_glob` to set these placeholders based on which files exist on disk).
 
         Args:
             key (str, Sequence[str]): identifier(s) for the template.
+            filter (bool, optional): If Set to True, will filter out any non-existent files.
+                If the return type is strings, non-existent entries will be empty strings.
+                If the return type is Path objects, non-existent entries will be None.
+                Note that the default behaviour is opposite from :meth:`get_mult_glob`.
             make_dir (bool, optional): If set to True, create the parent directory for each returned path.
 
         Returns:
             xarray.DataArray, xarray.Dataset: For a single key returns all possible paths in an xarray DataArray. 
                 For multiple keys it returns the combination of them in an xarray Dataset.
+                Each element of in the xarray is a `pathlib.Path` object if `FileTree.return_path` is True.
+                Otherwise the xarray will contain the paths as strings.
         """
         if isinstance(key, str):
-            paths = self.get_template(key).format_mult(self.placeholders)
+            paths = self.get_template(key).format_mult(self.placeholders, filter=filter)
             paths.name = key
             if make_dir:
                 for path in paths.data.flat:
-                    path.parent.mkdir(parents=True, exists_ok=True)
+                    if path is not None:
+                        path.parent.mkdir(parents=True, exists_ok=True)
+            if self.return_path:
+                return xarray.apply_ufunc(lambda p: None if p == "" else Path(p), paths, vectorize=True)
             return paths
         else:
-            return xarray.merge([self.get_mult(k) for k in key], join='exact')
-        
+            return xarray.merge([self.get_mult(k, filter=filter, make_dir=make_dir) for k in key], join='exact')
+
+    def get_mult_glob(self, key: Union[str, Sequence[str]]) -> Union[xarray.DataArray, xarray.Dataset]:
+        """Returns array of paths with all possible values filled in for the placeholders 
+
+        Singular placeholder values are filled into the template directly. 
+        For each placeholder with multiple values a dimension is added to the output array.
+        This dimension will have the name of the placeholder and labels corresponding to the possible values (see http://xarray.pydata.org/en/stable/).
+        The possible values for undefined placeholders will be determined by which files actually exist on disk.
+
+        The same result can be obtained by calling `self.update_glob(key).get_mult(key, filter=True)`.
+        However calling this method is more efficient, because it only has to check the disk for which files exist once.
+
+        Args:
+            key (str, Sequence[str]): identifier(s) for the template.
+
+        Returns:
+            xarray.DataArray, xarray.Dataset: For a single key returns all possible paths in an xarray DataArray. 
+                For multiple keys it returns the combination of them in an xarray Dataset.
+                Each element of in the xarray is a `pathlib.Path` object if `FileTree.return_path` is True.
+                Otherwise the xarray will contain the paths as strings.
+        """
+        if isinstance(key, str):
+            template = self.get_template(key)
+            matches = template.all_matches(self.placeholders)
+
+            new_placeholders = Placeholders(self.placeholders)
+            new_placeholders.update(template.get_all_placeholders(self.placeholders, matches=matches))
+
+            paths = template.format_mult(new_placeholders, filter=True, matches=matches)
+            paths.name = key
+            if self.return_path:
+                return paths
+            res = xarray.apply_ufunc(lambda p: "" if p is None else str(p), paths, vectorize=True)
+            return res
+        else:
+            return xarray.merge([self.get_mult_glob(k) for k in key], join='outer', fill_value=None if self.return_path else "")
+
 
     def fill(self, ) -> "FileTree":
         """Fills in singular placeholder values.
 
         Returns:
-            FileTree: new tree with singular placeholder values filled in and removed from the placeholder dict
+            FileTree: new tree with singular placeholder values filled into the templates and removed from the placeholder dict
         """
         new_templates = {}
         to_assign = dict(self._templates)
@@ -373,7 +455,7 @@ class FileTree:
             Generator[FileTree]: yields trees, where each placeholder only has a single possible value
         """
         for placeholders in self.placeholders.iter_over(placeholders):
-            yield FileTree(self._templates, placeholders)
+            yield FileTree(self._templates, placeholders, self.return_path)
 
     def iter(self, template: str, check_exists: bool=False) -> Generator["FileTree", None, None]:
         """Iterate over trees containng all possible values for template
@@ -385,7 +467,7 @@ class FileTree:
         Returns:
             Generator[FileTree]: yields trees, where each placeholder in given template only has a single possible value
         """
-        placeholders = self.get_template(template).placeholders()
+        placeholders = self.get_template(template).placeholders(valid=self.placeholders.keys())
         for tree in self.iter_vars(placeholders):
             if not check_exists or tree.get(template).exists:
                 yield tree
@@ -419,7 +501,7 @@ class FileTree:
         already_done = set()
         for _, t in top_level:
             if t not in already_done:
-                lines.append(t.as_string(self._templates, indentation=indentation))
+                lines.append(t.as_multi_line(self._templates, indentation=indentation))
                 already_done.add(t)
         return '\n\n'.join(lines)
 
