@@ -1,18 +1,31 @@
-from enum import unique
 from pathlib import Path
 import numpy as np
 import re
 import itertools
-from typing import Generator, List, Optional, Sequence, Set, Tuple, Dict, Iterator, Union
+from typing import Generator, List, Optional, Sequence, Set, Tuple, Dict, Iterator, Any
 import xarray
 import pandas as pd
-from xarray.core.utils import V
 from collections.abc import MutableMapping
-from itertools import product, combinations
-from glob import iglob
+from itertools import product, combinations, chain
+from glob import glob
 from collections import defaultdict
 from datetime import datetime
-from parse import parse, extract_format
+from parse import extract_format, compile
+from functools import cmp_to_key, lru_cache
+import pandas as pd
+import os
+
+
+def is_singular(value):
+    """Whether a value is singular or has multiple options.
+    """
+    if isinstance(value, str):
+        return True
+    try:
+        iter(value)
+        return False
+    except TypeError:
+        return True
 
 class Placeholders(MutableMapping):
     """Dictionary-like object containing the placeholder values.
@@ -24,6 +37,12 @@ class Placeholders(MutableMapping):
         self.linkages = {}
         self.update(dict(*args, **kwargs))
 
+    def copy(self):
+        p = Placeholders()
+        p.mapping = dict(self.mapping)
+        p.linkages = dict(self.linkages)
+        return p
+
     def __getitem__(self, key: str):
         actual_key = self.find_key(key)
         if actual_key is None:
@@ -90,9 +109,9 @@ class Placeholders(MutableMapping):
         Returns:
             Optional[str]: None if no value for the key is available, otherwise the key used to index the value
         """
-        if isinstance(key, tuple):
+        if not isinstance(key, str):
             key = frozenset(key)
-        if key in self.linkages:
+        elif key in self.linkages:
             return key
         if self.mapping.get(key, None) is not None:
             return key
@@ -118,7 +137,7 @@ class Placeholders(MutableMapping):
         single_placeholders = Placeholders()
         multi_placeholders = Placeholders()
         for name, value in self.mapping.items():
-            if isinstance(name, tuple) or np.asanyarray(value).ndim == 1:
+            if isinstance(name, frozenset) or not is_singular(value):
                 multi_placeholders[name] = value
             else:
                 single_placeholders[name] = value
@@ -142,17 +161,17 @@ class Placeholders(MutableMapping):
         iter_values = {}
         for key in actual_keys:
             if key not in unique_keys:
-                if isinstance(key, (frozenset, tuple)):  # linked placeholder
+                if isinstance(key, frozenset):  # linked placeholder
                     unique_keys.append(key)
-                    iter_values[key] = ({k: self[k][idx] for k in key} for idx in range(len(self[list(key)[0]])))
-                elif np.asanyarray(self[key]).ndim == 1:  # iterable placeholder
+                    iter_values[key] = [{k: self[k][idx] for k in key} for idx in range(len(self[list(key)[0]]))]
+                elif not is_singular(self[key]):  # iterable placeholder
                     unique_keys.append(key)
                     iter_values[key] = self[key]
 
         for values in product(*[iter_values[k] for k in unique_keys]):
             new_vars = Placeholders(self)
             for key, value in zip(unique_keys, values):
-                if isinstance(key, (tuple, frozenset)):
+                if isinstance(key, frozenset):
                     del new_vars[key]  # break the placeholders link
                     new_vars.update(value)
                 else:
@@ -161,15 +180,83 @@ class Placeholders(MutableMapping):
 
     def link(self, *keys):
         """
-        Link the placeholders represented by `keys`
+        Link the placeholders represented by `keys`.
+
+        When iterating over linked placeholders the i-th tree will contain the i-th element from all linked placeholders,
+        instead of the tree containing all possible combinations of placeholder values.
+
+        This can be thought of using `zip` for linked variables and `itertools.product` for unlinked ones.
         """
         self[keys] = [self[key] for key in keys]
 
     def unlink(self, *keys):
+        """
+        Unlink the placeholders represented by `keys`.
+
+        See :meth:`link` for how linking affects the iteration through placeholders with multiple values.
+
+        Raises a ValueError if the placeholders are not actually linked.
+        """
+        if keys not in self:
+            raise ValueError(f"{keys} were not linked, so cannot unlink them")
         new_vars = {k: self[k] for k in keys}
         del self[keys]
         self.update(new_vars)
-        
+
+class MyDataArray:
+    """Wrapper around xarray.DataArray for internal usage
+
+    It tries to delay creating the DataArray object as long as possible (as using them for small arrays is slow...)
+    """
+    def __init__(self, data, coords=None):
+        self.as_xarray = coords is None
+        if self.as_xarray:
+            assert isinstance(data, xarray.DataArray)
+            self.data_array = data
+        else:
+            self.data = data
+            self.coords = coords
+
+    def map(self, func) -> "MyDataArray":
+        if self.as_xarray:
+            return MyDataArray(xarray.apply_ufunc(func, self.data_array, vectorize=True))
+        else:
+            return MyDataArray(np.array([func(d) for d in self.data.flat]).reshape(self.data.shape), self.coords)
+
+    def to_xarray(self, ) -> xarray.DataArray:
+        if self.as_xarray:
+            return self.data_array
+        else:
+            return xarray.DataArray(self.data, [_to_index(name, values) for name, values in self.coords])
+
+    @staticmethod
+    def concat(parts, new_index) -> "MyDataArray":
+        if len(parts) == 0:
+            return MyDataArray(np.array([]), [])
+        to_xarray = (
+            any(p.as_xarray for p in parts) or
+            any(
+                len(p.coords) != len(parts[0].coords) or
+                any(np.all(name1 != name2) for (name1, _), (name2, _) in zip(p.coords, parts[0].coords))
+                for p in parts
+            )
+        )
+        if to_xarray:
+            return MyDataArray(xarray.concat([p.to_xarray() for p in parts], _to_index(*new_index)))
+        else:
+            new_data = np.stack([p.data for p in parts], axis=0)
+            new_coords = list(parts[0].coords)
+            new_coords.insert(0, new_index)
+            return MyDataArray(new_data, new_coords)
+
+
+def _to_index(name, values):
+    if isinstance(name, str):
+        return pd.Index(values, name=name)
+    else:
+        return pd.MultiIndex.from_tuples(values, names=name)
+
+
 
 class Template:
     def __init__(self, parent: "Template", unique_part: str):
@@ -184,8 +271,14 @@ class Template:
             return Path(self.unique_part)
         return self.parent.as_path.joinpath(self.unique_part)
 
+    @property
+    def as_string(self, ):
+        if self.parent is None:
+            return str(self.unique_part)
+        return os.path.join(self.parent.as_string, str(self.unique_part))
+
     def __str__(self, ):
-        return f"Template({str(self.as_path)})"
+        return f"Template({self.as_string})"
 
     def children(self, templates: Sequence["Template"]) -> List["Template"]:
         """From a sequence of templates find the children
@@ -195,17 +288,17 @@ class Template:
         """
         return [t for t in templates if t.parent is self]
 
-    def as_string(self, other_templates: Dict[str, "Template"], indentation=4) -> str:
+    def as_multi_line(self, other_templates: Dict[str, "Template"], indentation=4) -> str:
         """Generates a string describing this and child templates as 
 
         Args:
             other_templates (Dict[str, Template]): templates including all the child templates and itself
-            indentiation (int, optional): number of spaces to use as indentation. Defaults to 4
+            indentation (int, optional): number of spaces to use as indentation. Defaults to 4
 
         Returns:
             str: multi-line string that can be processed by :meth:`file_tree.FileTree.read`
         """
-        result = self._as_string_helper(other_templates, indentation)
+        result = self._as_multi_line_helper(other_templates, indentation)
 
         is_top_level = other_templates[""] is self
         if not is_top_level and self.parent is None:
@@ -213,7 +306,7 @@ class Template:
         else:
             return result
 
-    def _as_string_helper(self, other_templates: Dict[str, "Template"], indentation=4, _current_indentation=0) -> str:
+    def _as_multi_line_helper(self, other_templates: Dict[str, "Template"], indentation=4, _current_indentation=0) -> str:
         leaves = []
         branches = []
         for t in sorted(self.children(other_templates.values()), key=lambda t: t.unique_part):
@@ -248,23 +341,23 @@ class Template:
         already_done = set()
         for t in leaves + branches:
             if t not in already_done:
-                lines.append(t._as_string_helper(other_templates, indentation, indentation + _current_indentation))
+                lines.append(t._as_multi_line_helper(other_templates, indentation, indentation + _current_indentation))
                 already_done.add(t)
         return '\n'.join(lines)
 
     @property
     def _parts(self, ):
-        return TemplateParts.parse(str(self.as_path))
+        return TemplateParts.parse(self.as_string)
 
-    def placeholders(self, ) -> List[str]:
+    def placeholders(self, valid=None) -> List[str]:
         """Returns a list of the placeholder names
 
         Returns:
             List[str]: placeholder names in order that they appear in the template
         """
-        return self._parts.ordered_placeholders()
+        return self._parts.ordered_placeholders(valid)
 
-    def format_single(self, placeholders: Placeholders, check=True) -> Path:
+    def format_single(self, placeholders: Placeholders, check=True) -> str:
         """Formats the template with the placeholders filled in
 
         Only placeholders with a single value are considered.
@@ -277,40 +370,45 @@ class Template:
             KeyError: if any placeholder is missing
 
         Returns:
-            Path: filled in template
+            str: filled in template
         """
         single_placeholders, _ = placeholders.split()
-        res = self._parts.resolve(single_placeholders)
-        if res.ndim > 0:
-            raise KeyError(f"Multiple placeholder values provided for {res.coords}")
-        template: TemplateParts = res.data[()]
+        template = self._parts.fill_single_placeholders(single_placeholders).remove_optionals()
         if check:
             unfilled = template.required_placeholders()
             if len(unfilled) > 0:
                 raise KeyError(f"Missing placeholder values for {unfilled}")
-        return Path(str(template))
+        return str(template)
 
-    def format_mult(self, placeholders: Placeholders, check=False) -> xarray.DataArray:
-        """Replaces placeholder with the provided placeholders
+    def format_mult(self, placeholders: Placeholders, check=False, filter=False, matches=None) -> xarray.DataArray:
+        """Replaces placeholders in template with the provided placeholder values
 
         Args:
             placeholders (Placeholders): mapping from placeholder names to single or multiple vaalues
             check (bool): skip check for missing placeholders if set to True
+            filter (bool): filter out non-existing files if set to True
 
         Raises:
             KeyError: if any placeholder is missing
 
         Returns:
-            xarray.DataArray: array with possible resolved paths
+            xarray.DataArray: array with possible resolved paths.
+                If `filter` is set to True the non-existent paths are replaced by None
         """
-        resolved = self._parts.resolve(placeholders)
+        parts = self._parts
+        resolved = parts.resolve(placeholders)
         if check:
             for template in resolved.data.flatten():
                 unfilled = template.required_placeholders()
                 if len(unfilled) > 0:
                     raise KeyError(f"Missing placeholder values for {unfilled}")
-        paths = xarray.apply_ufunc(lambda t: Path(str(t)), resolved, vectorize=True)
-        return paths
+        paths = resolved.map(lambda t: str(t))
+        if not filter:
+            return paths.to_xarray()
+        placeholder_dict = dict(placeholders)
+        path_matches = [str(parts.fill_single_placeholders(Placeholders({**placeholder_dict, **match})).remove_optionals())
+                        for match in (self.all_matches(placeholders) if matches is None else matches)]
+        return paths.map(lambda p: p if p in path_matches else '').to_xarray()
 
     def optional_placeholders(self, ) -> Set[str]:
         """Finds all placeholders that are only within optional blocks (i.e., they do not require a value)
@@ -358,32 +456,44 @@ class Template:
         updated = ''.join([str(p.add_precursor(text)) for p in parts])
         return Template(self.parent, updated)
 
-    def get_all_placeholders(self, placeholders: Placeholders) -> Placeholders:
+    def get_all_placeholders(self, placeholders: Placeholders, matches=None) -> Placeholders:
         """Fill placeholders with possible values based on what is available on disk
 
         Args:
             placeholders (Placeholders): New values for undefined placeholders in template
         """
-        single_vars, multi_vars = placeholders.split()
-        filled_template: TemplateParts = self._parts.fill_known(single_vars).data[()]
-        optionals = self.optional_placeholders()
         undefined = defaultdict(set)
-        for to_glob, templates in filled_template.all_optionals().items():
-            for fn in iglob(to_glob):
-                for t in templates:
-                    try:
-                        vars = t.parse_filename(fn)
-                    except ValueError:
-                        continue
-                    for var_name in optionals:
-                        if var_name not in vars:
-                            vars[var_name] = None
-                    if not all(name not in multi_vars or value in multi_vars[name] for name, value in vars.items()):
-                        continue
-                    for name, value in vars.items():
-                        if name not in placeholders:
-                            undefined[name].add(value)
-        return Placeholders(undefined)
+        for match in self.all_matches(placeholders) if matches is None else matches:
+            for name, value in match.items():
+                if name not in placeholders:
+                    undefined[name].add(value)
+
+        def cmp(item1, item2):
+            if item1 is None:
+                return -1
+            if item2 is None:
+                return 1
+            if item1 < item2:
+                return -1
+            if item1 > item2:
+                return 1
+            return 0
+
+        return Placeholders({k: sorted(v, key=cmp_to_key(cmp)) for k, v in undefined.items()})
+
+    def all_matches(self, placeholders: Placeholders):
+        """Returns a sequence of all possible variable values matching existing files on disk
+
+        Only variable values matching existing placeholder values are returned (undefined placeholders are unconstrained).
+        """
+        single_vars, multi_vars = placeholders.split()
+        filled_template: TemplateParts = self._parts.fill_single_placeholders(single_vars)
+        res = []
+        for match in filled_template.all_matches():
+            if not all(name not in multi_vars or value in multi_vars[name] for name, value in match.items()):
+                continue
+            res.append(match)
+        return res
 
 
 def extract_placeholders(template, filename, known_vars=None):
@@ -413,7 +523,7 @@ class Part:
         """
         Fills in the given placeholders
         """
-        return [self]
+        return (self, )
 
     def optional_placeholders(self, ) -> Set["Part"]:
         """
@@ -433,7 +543,7 @@ class Part:
         """
         return False
 
-    def append_placeholders(self, placeholders: List[str]):
+    def append_placeholders(self, placeholders: List[str], valid=None):
         """
         Appends the placeholders in this part to the provided list in order
         """
@@ -451,6 +561,9 @@ class Part:
         """
         return [self]
 
+    def remove_precursors(self, placeholders=None):
+        return self
+
 class Literal(Part):
     def __init__(self, text: str):
         """
@@ -466,6 +579,11 @@ class Literal(Part):
         """
         return self.text
 
+    def __eq__(self, other):
+        if not isinstance(other, Literal):
+            return NotImplemented
+        return self.text == other.text
+
 
 class Required(Part):
     def __init__(self, var_name, var_formatting=None):
@@ -492,7 +610,7 @@ class Required(Part):
     def fill_single_placeholders(self, placeholders: Placeholders, ignore_type=False):
         value = placeholders.get(self.var_name, None)
         if value is None:
-            return [self]
+            return (self, )
         else:
             if not ignore_type and len(self.var_formatting) > 0:
                 format_type = extract_format(self.var_formatting, [])["type"]
@@ -502,13 +620,17 @@ class Required(Part):
                     value = float(value)
                 elif format_type in ['t' + ft for ft in 'iegachs']:
                     value = datetime(value)
-            as_string = ('{' + self.var_name + '}') if ignore_type else str(self)
-            return TemplateParts.parse(as_string.format(**{self.var_name: value})).parts
+            res = TemplateParts.parse(format(value, '' if ignore_type else self.var_formatting))
+            if len(res.parts) == 1:
+                return res.parts
+            return res.fill_single_placeholders(placeholders, ignore_type=ignore_type).parts
 
     def required_placeholders(self, ):
         return {self.var_name}
 
-    def append_placeholders(self, placeholders):
+    def append_placeholders(self, placeholders, valid=None):
+        if valid is not None and self.var_name not in valid:
+            raise ValueError(f"Placholder {self.var_name} is not defined")
         placeholders.append(self.var_name)
 
     def add_precursor(self, text: str) -> "Required":
@@ -516,6 +638,18 @@ class Required(Part):
         """
         return Required(text + self.var_name, self.var_formatting)
 
+    def remove_precursors(self, placeholders=None):
+        if placeholders is None:
+            new_name = self.var_name.split('/')[-1]
+        else:
+            key =  placeholders.find_key(self.var_name)
+            new_name = self.var_name if key is None else key
+        return Required(new_name, self.var_formatting)
+
+    def __eq__(self, other):
+        if not isinstance(other, Required):
+            return NotImplemented
+        return (self.var_name == other.var_name) & (self.var_formatting == other.var_formatting)
 
 class OptionalPart(Part):
     def __init__(self, sub_template: "TemplateParts"):
@@ -534,8 +668,8 @@ class OptionalPart(Part):
     def fill_single_placeholders(self, placeholders: Placeholders, ignore_type=False):
         new_opt = self.sub_template.fill_single_placeholders(placeholders, ignore_type=ignore_type)
         if len(new_opt.required_placeholders()) == 0:
-            return TemplateParts.parse(str(new_opt)).parts
-        return [OptionalPart(new_opt)]
+            return (Literal(str(new_opt)), )
+        return (OptionalPart(new_opt), )
 
     def optional_placeholders(self, ):
         return self.sub_template.required_placeholders()
@@ -545,8 +679,11 @@ class OptionalPart(Part):
             return True
         return len(self.optional_placeholders().intersection(placeholders)) > 0
 
-    def append_placeholders(self, placeholders):
-        placeholders.extend(self.sub_template.ordered_placeholders())
+    def append_placeholders(self, placeholders, valid=None):
+        try:
+            placeholders.extend(self.sub_template.ordered_placeholders(valid=valid))
+        except ValueError:
+            pass
 
     def add_precursor(self, text: str) -> "OptionalPart":
         return OptionalPart(TemplateParts([p.add_precursor(text) for p in self.sub_template.parts]))
@@ -556,23 +693,35 @@ class OptionalPart(Part):
 
         Removes any optional parts, whose placeholders are not in `placeholder_names`.
         """
-        if len(self.optional_placeholders().intersection(placeholder_names)) > 0:
+        if len(self.optional_placeholders().difference(placeholder_names)) > 0:
             return []
         return self.sub_template.parts
 
+    def remove_precursors(self, placeholders=None):
+        return OptionalPart(self.sub_template.remove_precursors(placeholders))
+
+    def __eq__(self, other):
+        if not isinstance(other, OptionalPart):
+            return NotImplemented
+        return (self.sub_template == other.sub_template)
+
 
 class TemplateParts:
     """
     The parts of a larger template
     """
+    optional_re = re.compile(r'(\[.*?\])')
+    requires_re = re.compile(r'(\{.*?\})')
+
     def __init__(self, parts: Sequence[Part]):
         if isinstance(parts, str):
             raise ValueError("Input to Template should be a sequence of parts; " +
                              "did you mean to call `TemplateParts.parse` instead?")
         self.parts = tuple(parts)
 
-    @classmethod
-    def parse(cls, text: str) -> "TemplateParts":
+    @staticmethod
+    @lru_cache(1000)
+    def parse(text: str) -> "TemplateParts":
         """Parses a template string into its constituent parts
 
         Raises:
@@ -583,13 +732,13 @@ class TemplateParts:
         """
 
         parts = []
-        for optional_parts in re.split(r'(\[.*?\])', text):
+        for optional_parts in TemplateParts.optional_re.split(text):
             if len(optional_parts) > 0 and optional_parts[0] == '[' and optional_parts[-1] == ']':
                 if '[' in optional_parts[1:-1] or ']' in optional_parts[1:-1]:
                     raise ValueError(f'Can not parse {text}, because unmatching square brackets were found')
                 parts.append(OptionalPart(TemplateParts.parse(optional_parts[1:-1])))
             else:
-                for required_parts in re.split(r'(\{.*?\})', optional_parts):
+                for required_parts in TemplateParts.requires_re.split(optional_parts):
                     if len(required_parts) > 0 and required_parts[0] == '{' and required_parts[-1] == '}':
                         if ':' in required_parts:
                             var_name, var_type = required_parts[1:-1].split(':')
@@ -604,7 +753,7 @@ class TemplateParts:
         """
         Returns the template as a string
         """
-        return str(Path(''.join([str(p) for p in self.parts])))
+        return os.path.normpath(''.join([str(p) for p in self.parts]))
 
     def optional_placeholders(self, ) -> Set[str]:
         """Set of optional placeholders
@@ -621,41 +770,58 @@ class TemplateParts:
             return set()
         return set.union(*[p.required_placeholders() for p in self.parts])
 
-    def ordered_placeholders(self, ) -> Tuple[str]:
+    def ordered_placeholders(self, valid=None) -> Tuple[str]:
         """Sequence of all placeholders in order (can contain duplicates)
         """
         ordered_vars = []
         for p in self.parts:
-            p.append_placeholders(ordered_vars)
+            p.append_placeholders(ordered_vars, valid=valid)
         return ordered_vars
 
-    def fill_known(self, placeholders: Placeholders, ignore_type=False) -> xarray.DataArray:
+    def fill_known(self, placeholders: Placeholders, ignore_type=False) -> MyDataArray:
         """Fill in the known placeholders
 
         Any optional parts, where all placeholders have been filled will be automatically replaced
         """
         single, multi = placeholders.split()
+        return self.remove_precursors(placeholders)._fill_known_helper(single, multi, ignore_type=ignore_type)
+
+    def _fill_known_helper(self, single: Placeholders, multi: Placeholders, ignore_type=False) -> MyDataArray:
         new_template = self.fill_single_placeholders(single, ignore_type=ignore_type)
         for name in new_template.ordered_placeholders():
-            if name in multi:
-                values = np.asanyarray(multi[name])
-                templates = [new_template.fill_known(Placeholders({**placeholders, **{name: value}}), ignore_type=ignore_type) for value in values]
-                return xarray.concat(templates, pd.Index(values, name=name))
-        return xarray.DataArray(new_template)
+            use_name = multi.find_key(name)
+            if use_name is None:
+                continue
+            new_multi = multi.copy()
+            if use_name in multi.linkages:
+                values = multi[multi.linkages[use_name]]
+                keys = tuple(sorted(values.keys()))
+                index = (keys, zip(*[values[k] for k in keys]))
+                del new_multi[new_multi.linkages[use_name]]
+            else:
+                values = {use_name: list(multi[name])}
+                index = (use_name, values[use_name])
+                del new_multi[use_name]
+            assert use_name is not None
 
-    def fill_single_placeholders(self, placeholders: Placeholders, ignore_type=False):
+            parts = []
+            new_single = single.copy()
+            for idx in range(len(values[use_name])):
+                new_vals = {n: v[idx] for n, v in values.items()}
+                new_single.mapping.update(new_vals)
+                parts.append(new_template._fill_known_helper(new_single, new_multi, ignore_type=ignore_type))
+
+            return MyDataArray.concat(parts, index)
+        return MyDataArray(np.array(new_template), [])
+
+    def fill_single_placeholders(self, placeholders: Placeholders, ignore_type=False) -> "TemplateParts":
         """
-        Helper method for :meth:`fill_known`
+        Fills in placeholders with singular values
+
+        Assumes that all placeholders are in fact singular
         """
-        current = self
-        prev = ''
-        while str(current) != prev:
-            prev = str(current)
-            res = []
-            for p in current.parts:
-                res.extend(p.fill_single_placeholders(placeholders, ignore_type=ignore_type))
-            current = TemplateParts(res)
-        return current
+        res = [p.fill_single_placeholders(placeholders, ignore_type=ignore_type) for p in self.parts]
+        return TemplateParts(list(chain(*res)))
 
     def remove_optionals(self, optionals=None) -> "TemplateParts":
         """
@@ -663,27 +829,38 @@ class TemplateParts:
         """
         return TemplateParts([p for p in self.parts if not p.contains_optionals(optionals)])
 
-    def all_optionals(self, ) -> Dict[str, List["TemplateParts"]]:
-        """Yields all possible templates depending on which optional placeholders are defined
+    def all_matches(self, ) -> List[Dict[str, Any]]:
+        """Finds all potential matches to existing templates
 
-        None of the resulting templates will have optional placeholders (i.e., square brackets)
+        Returns a list with the possible combination of values for the placeholders.
         """
         required = self.required_placeholders()
         optional = self.optional_placeholders()
-        to_glob = defaultdict(list)
+        matches = []
+        already_globbed = {}
         for defined_optionals in [c for n in range(len(optional) + 1) for c in combinations(optional, n)]:
-            new_glob = str(self.resolve(Placeholders({req: '*' for req in required}, **{opt: '*' for opt in defined_optionals}), ignore_type=True).data[()])
+            new_glob = str(self.fill_single_placeholders({**{req: '*' for req in required}, **{opt: '*' for opt in defined_optionals}}, ignore_type=True).remove_optionals())
             while '**' in new_glob:
                 new_glob = new_glob.replace('**', '*')
-            if new_glob not in to_glob:
-                res = []
-                vars = required.union(defined_optionals)
-                for p in self.parts:
-                    res.extend(p.for_defined(vars))
-                to_glob[new_glob].append(TemplateParts(res))
-        return to_glob
+            if new_glob not in already_globbed:
+                already_globbed[new_glob] = glob(new_glob)
+            res = []
+            vars = required.union(defined_optionals)
+            for p in self.parts:
+                res.extend(p.for_defined(vars))
+            parser = TemplateParts(res).get_parser()
+            for fn in already_globbed[new_glob]:
+                try:
+                    placeholders = parser(fn)
+                except ValueError:
+                    continue
+                for var_name in optional:
+                    if var_name not in placeholders:
+                        placeholders[var_name] = None
+                matches.append(placeholders)
+        return matches
 
-    def resolve(self, placeholders, ignore_type=False) -> str:
+    def resolve(self, placeholders, ignore_type=False) -> MyDataArray:
         """
         Resolves the template given a set of placeholders
 
@@ -691,7 +868,7 @@ class TemplateParts:
         :param ignore_type: if True, ignore the type formatting when filling in placeholders
         :return: cleaned string
         """
-        return xarray.apply_ufunc(lambda t: t.remove_optionals(), self.fill_known(placeholders, ignore_type=ignore_type), vectorize=True)
+        return self.fill_known(placeholders, ignore_type=ignore_type).map(lambda t: t.remove_optionals())
 
     def optional_subsets(self, ) -> Iterator["TemplateParts"]:
         """
@@ -771,13 +948,30 @@ class TemplateParts:
                 raise KeyError("Multiple equivalent ways found to parse {} using {}".format(filename, template))
         return best
 
-    def parse_filename(self, filename):
+    def get_parser(self):
         if any(isinstance(p, OptionalPart) for p in self.parts):
             raise ValueError("Can not parse filename when there are optional parts in the template")
-        result = parse(str(self), filename, case_sensitive=True)
-        if result is None:
-            raise ValueError(f"template string ({str(self)}) does not mach filename ({filename})")
-        named = result.named
-        if any(isinstance(value, str) and '/' in value for value in named.values()):
-            raise ValueError("Placeholder can not span directories")
-        return named
+        parser = compile(str(self.remove_precursors()), case_sensitive=True)
+
+        def parse_filename(filename):
+            result = parser.parse(filename)
+            if result is None:
+                raise ValueError(f"template string ({str(self)}) does not mach filename ({filename})")
+            named = result.named
+            if any(isinstance(value, str) and '/' in value for value in named.values()):
+                raise ValueError("Placeholder can not span directories")
+            return named
+        
+        return parse_filename
+
+    def remove_precursors(self, placeholders=None):
+        """Replaces keys to those existing in the placeholders
+
+        If no placeholders provided all precursors are removed
+        """
+        return TemplateParts([p.remove_precursors(placeholders) for p in self.parts])
+
+    def __eq__(self, other):
+        if not isinstance(other, TemplateParts):
+            return NotImplemented
+        return (len(self.parts) == len(other.parts)) and all(p1 == p2 for p1, p2 in zip(self.parts, other.parts))
\ No newline at end of file
